@@ -107,6 +107,19 @@ class TestIndexThings:
         if m:  # only server-side mapp returns messages
             assert "not/exists" in m
 
+    def test_pypi_index_attributes(self, mapp):
+        mapp.login_root()
+        data = mapp.getjson("/root/pypi")
+        res = data["result"]
+        assert res["volatile"] == False
+        assert not res["uploadtrigger_jenkins"]
+        assert res["acl_upload"] == ["root"]
+
+    def test_pypi_not_modifiable(self, mapp):
+        mapp.login_root()
+        res = mapp.getjson("/root/pypi")["result"]
+        mapp.modify_index("root/pypi", res, code=403)
+
     def test_create_index_base_empty(self, mapp):
         indexconfig = dict(bases=())
         mapp.login_root()
@@ -134,7 +147,7 @@ class TestIndexThings:
         mapp.login("root", "")
         mapp.change_password("root", "asd")
         mapp.create_and_login_user("newuser1")
-        mapp.create_index("root/test2", code=401)
+        mapp.create_index("root/test2", code=403)
 
     def test_create_index_and_acls(self, mapp):
         username = "newuser2"
@@ -164,14 +177,13 @@ class TestIndexThings:
         mapp.create_and_login_user("cuser1")
         data = mapp.getjson("/cuser1")
         assert data["type"] == "userconfig"
-        data = mapp.getjson("/cuser1/")
-        assert data["type"] == "list:indexconfig"
 
     def test_create_index_and_config_gets(self, mapp):
         mapp.create_and_login_user("cuser2")
         mapp.create_index("dev")
-        assert mapp.getjson("/cuser2/dev")["type"] == "indexconfig"
-        assert mapp.getjson("/cuser2/dev/")["type"] == "list:projectconfig"
+        res =  mapp.getjson("/cuser2/dev")
+        assert res["type"] == "indexconfig"
+        assert res["result"]["projects"] == []
 
     def test_non_volatile_cannot_be_deleted(self, mapp):
         mapp.create_and_login_user("cuser4")
@@ -179,3 +191,12 @@ class TestIndexThings:
         mapp.delete_index("dev", code=403)
         mapp.delete_user("cuser4", code=403)
 
+    def test_custom_data(self, mapp):
+        mapp.create_and_login_user("cuser5")
+        mapp.create_index("dev")
+        mapp.use("cuser5/dev")
+        res = mapp.getjson("/cuser5/dev")
+        assert "custom_data" not in res["result"]
+        mapp.set_custom_data("foo")
+        res = mapp.getjson("/cuser5/dev")
+        assert res["result"]["custom_data"] == "foo"
